from config import get_logger, get_rich_console
from yaml_config import find_and_parse_yaml_services
from typing import Dict, Union

console = get_rich_console()
logger = get_logger(__name__)


class ResourceNodeRegistry:
    _registry = {}
    _yaml_config_registry = {}

    def __init__(self) -> None:
        yaml_services = find_and_parse_yaml_services()
        for yaml_service in yaml_services:
            service_name = yaml_service.service_name
            if service_name not in self._yaml_config_registry:
                self._yaml_config_registry[service_name] = {}

            for yaml_r_node_config in yaml_service.resource_nodes:
                rn_name = yaml_r_node_config.resource_node_name
                self._yaml_config_registry[service_name][rn_name] = yaml_r_node_config
                logger.debug(f"YamlRegistry: Registered {service_name}.{rn_name}")

    def search_yaml_config_registry(self, service_name, resource_node_name):
        return self._yaml_config_registry.get(service_name, {}).get(
            resource_node_name, None
        )

    def register_class(
        self, cls: "ResourceNode", service_name: str = None, name: str = None
    ) -> None:
        """Register a subclass of ResourceNode with the given `service_name` and `name`.
        When a ResourceNode object is created, registered classes will be selected as the
        ResourceNode class. This allows overriding ability to each service/resource.

        Args:
            cls (ResourceNode): A subclass of ResourceNode
            service_name (str, optional): ServiceNode name, must be provided. Defaults to None.
            name (str, optional): ResourceNode name, must be provided. Defaults to None.


        """
        # TODO: check if service_name is correct get_available_service_names
        service_dict = self._registry.get(service_name, False)
        if not service_dict:
            self._registry[service_name] = {name: cls}
        else:
            custom_resource_node = service_dict.get(name, False)
            if custom_resource_node:
                return False
            else:
                self._registry[service_name][name] = cls

    def search_registry_for_service(
        cls, service_name: str
    ) -> Dict[str, "ResourceNode"]:
        """Returns services registered custom ResourceNode subclasses.

        Returns:
            Dict[str, 'ResourceNode']: ResourceNode name to custom Subclasses mapping for the service.
        """
        return cls._registry.get(service_name, {})

    def find_custom_class_for_resource_node(
        cls, service_name: str, node_name: str
    ) -> Union["ResourceNode", bool]:
        """Try to find the ResourceNode's custom subclasses registered under the Service.

        Returns:
            Union["ResourceNode", bool] : False if no custom class is found, else the custom class
        """
        services_custom_nodes = cls.search_registry_for_service(service_name)
        for custom_cls_name, _cls_obj in services_custom_nodes.items():
            # if compare_two_camel_case_words(custom_cls_name, node_name):
            if custom_cls_name.lower() == node_name.lower():
                return _cls_obj

        return False



from typing import List, Dict, Optional, Any
from pydantic import BaseModel, validator


class YamlRelation(BaseModel):
    service_name: str
    resource_node_name: str
    operation_name: str
    required_shape_name: str
    target_shape_name: str
    target_shape_type: str
    target_path: str


class YamlComplementApiParameterAction(BaseModel):
    """Defines the `add` and `remove` options for complement_api_parameters in YamlResourceNodeOperation"""
    _ACTION_TYPES = ("add", "remove")

    action: str
    data: Optional[Dict[str, Any]]
    keys: Optional[List[str]]

    @validator("action")
    def action_must_be_valid(cls, v):
        """Validates the action field. Must be 'add' or 'remove'"""
        if v not in cls._ACTION_TYPES:
            raise ValueError("action must be 'add' or 'remove'")
        return v


class YamlResourceNodeOperation(BaseModel):
    """Defines the customizations for a specific operation in a Resource Node."""
    operation_name: str
    jmespath_selector: Optional[str]
    complement_api_parameters: Optional[List[YamlComplementApiParameterAction]]
    explicit_relations: Optional[List[YamlRelation]]
    override_api_parameters: Optional[List[Dict[str, Any]]]
    pagination_token_mapping: Optional[Dict[str, str]]


class YamlServiceResourceNode(BaseModel):
    """Defines a Resource Node in a Service. It can have extra_relations and list of operations"""
    resource_node_name: str
    extra_relations: Optional[List[YamlRelation]]
    operations: Optional[List[YamlResourceNodeOperation]]


class YamlService(BaseModel):
    """Defines a Service in a Yaml file. It can have multiple Resource Nodes"""
    service_name: str
    resource_nodes: Optional[List[YamlServiceResourceNode]]

    # @validator("resource_node")
    # def validate_resource_node(cls, value):
    #     if not value.extra_relations and not value.operations:
    #         raise ValueError(
    #             "Either extra_relations or operations must be defined in resource_node"
    #         )
    #     return value



import logging
from rich.logging import RichHandler
from rich.console import Console
import os
from pathlib import Path

HOME_DIR = os.path.expanduser('~')

# Defaults to ~/.balcony/
BALCONY_CONFIG_DIR = os.getenv('BALCONY_CONFIG_DIR', os.path.join(HOME_DIR, '.balcony'))

# Defaults to ~/.balcony/relations/
BALCONY_RELATIONS_DIR = os.getenv('BALCONY_RELATIONS_DIR', os.path.join(BALCONY_CONFIG_DIR, 'relations'))

# create the relations directory if not exists
Path(BALCONY_RELATIONS_DIR).mkdir(parents=True, exist_ok=True)

LOG_LEVEL = 'INFO'

# YamlResourceNode customization parameters
YAML_IGNORE_PREFIX = "_"
YAML_SERVICES_DIRECTORY = Path(__file__).parent / "custom_yamls"

_console = Console(color_system="auto", markup=True, )
_balcony_loggers = []


def clear_relations_cache() -> None:
    """Removes the cached relations of services in the  `$BALCONY_RELATIONS_DIR/<service>.json` files."""
    relation_files = os.listdir(BALCONY_RELATIONS_DIR)
    deleted_filenames = []
    for rel_file in relation_files:
        if rel_file.endswith('.json'):
            rel_file_abs_path = os.path.join(BALCONY_RELATIONS_DIR, rel_file)
            if os.path.exists(rel_file_abs_path):
                os.remove(rel_file_abs_path)
                deleted_filenames.append(rel_file_abs_path)
    return deleted_filenames


def get_rich_console() -> Console:
    """Returns the global defined `rich.console.Console` object for common use.

    Returns:
        rich.console.Console: rich console object
    """
    return _console


def supress_other_module_logs() -> None:
    """Sets Python modules log levels to `logging.CRITICAL` to supress them."""
    _supress_module_names = ('boto', 'urllib3', 's3transfer', 'boto3', 'botocore', 'nose')
    for _logger_name in logging.Logger.manager.loggerDict.keys():
        if _logger_name in _supress_module_names:
            logging.getLogger(_logger_name).setLevel(logging.CRITICAL)


def set_log_level_at_runtime(log_level: str):
    """Sets the log level of loggers that balcony uses at runtime.

    Args:
        log_level (str): Log level string: INFO, WARNING, DEBUG etc.
    """
    for _logger in _balcony_loggers:
        _logger.setLevel(log_level)
        # for handler in _logger.handlers:
        #     if isinstance(handler, type(logging.StreamHandler())):
        #         handler.setLevel(log_level)
        #         _logger.debug('Debug logging enabled')


def get_logger(name: str) -> logging.Logger:
    """Logger creation with RichHandler.

    Args:
        name (str): Logger name. Usually given `__name__`.

    Returns:
        logging.Logger: Logger obj
    """
    supress_other_module_logs()
    logging.basicConfig(
        level=LOG_LEVEL,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(markup=True, console=_console)]
    )

    _logger = logging.getLogger(name)
    # keep track of the created loggers
    _balcony_loggers.append(_logger)
    return _logger



from nodes import ResourceNode
from config import get_logger

logger = get_logger(__name__)



class Parameter(ResourceNode, service_name="ssm", name="Parameter"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "service_name": "ssm",
            "resource_node_name": "Parameter",
            "operation_name": "GetParameters",
            "required_shape_name": "Name",
            "target_shape_name": "Name",
            "target_shape_type": "string",
            "target_path": "Parameters[*].Name",
        }]
        
class ParameterHistory(ResourceNode, service_name="ssm", name="ParameterHistory"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def get_operations_relations(self, operation_name: str):
        r = super().get_operations_relations(operation_name)
        return [{
            "service_name": "ssm",
            "resource_node_name": "Parameter",
            "operation_name": "DescribeParameters",
            "required_shape_name": "Name",
            "target_shape_name": "Name",
            "target_shape_type": "string",
            "target_path": "Parameters[*].Name",
        }], None
    



import json
import textwrap
from utils import get_all_available_services, ifind_similar_names_in_list
from config import (
    get_logger,
    get_rich_console,
    set_log_level_at_runtime,
    clear_relations_cache,
)
# required for loading custom resource nodes into registry
from custom_nodes import *  # noqa
from aws import BalconyAWS

import typer
import jmespath
from typing import Optional, List, Dict, Generator
from rich.columns import Columns
from rich.panel import Panel
import logging
import boto3
from pathlib import Path


console = get_rich_console()
logger = get_logger(__name__)
session = boto3.session.Session()
balcony_aws = BalconyAWS(session)
app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False)


@app.callback()
def _main_app_callback(
    debug: bool = typer.Option(False, "--debug", "-d", help="Enable debug messages."),
):
    if debug:
        set_log_level_at_runtime(logging.DEBUG)


def save_dict_to_output_file(output_filepath: str, data: Dict):
    with open(output_filepath, "w") as f:
        json.dump(data, f, indent=2, default=str)


def save_str_list_to_output_file(output_filepath: str, lines: List[str]):
    with open(output_filepath, "w") as f:
        f.writelines([line + "\n" for line in lines])


def _complete_service_name(incomplete: str) -> Generator[str, None, None]:
    """Typer autocompletion function. Finds matching service names.

    Args:
        incomplete (str): Service names that're starting with this string.

    Yields:
        _type_: _description_
    """
    service_names = get_all_available_services(session)
    if not incomplete:
        for name in service_names:
            yield name
    else:
        for name in service_names:
            if name.startswith(incomplete):
                yield name


def _complete_resource_node_name(ctx: typer.Context, incomplete: str) -> Generator[str, None, None]:
    service = ctx.params.get("service", False)
    if not service:
        return []
    service_node = balcony_aws.get_service_node(service)
    if not service_node:
        return []
    resource_nodes = service_node.get_resource_nodes()
    if not resource_nodes:
        return []

    resource_node_names = []

    for _rn in resource_nodes:
        _rn_name = _rn.name
        if incomplete:
            if _rn_name.lower().startswith(incomplete.lower()):
                resource_node_names.append(_rn_name)
        else:
            resource_node_names.append(_rn_name)
    return resource_node_names


def _complete_operation_type(ctx: typer.Context) -> List[str]:
    service = ctx.params.get("service", False)
    resource_node = ctx.params.get("resource_node", False)
    if not service or not resource_node:
        return []
    service_node = balcony_aws.get_service_node(service)
    if not service_node:
        return []
    resource_node_obj = service_node.get_resource_node_by_name(resource_node)
    if not resource_node_obj:
        return []
    operation_types_and_names = resource_node_obj.get_operation_types_and_names()
    return list(operation_types_and_names.keys())


def _list_service_or_resource(
    service: Optional[str] = False,
    resource_node: Optional[str] = False,
    screen_pager: Optional[bool] = False,
) -> None:

    available_service_names = get_all_available_services(session)
    if not service and not resource_node:
        # nothing is given
        console.print(Columns(available_service_names, equal=True, expand=True))
        return available_service_names

    if service and not resource_node:
        # we only have service name

        if service not in available_service_names:
            similar_service_names = ifind_similar_names_in_list(
                service, available_service_names
            )
            if similar_service_names:
                console.print("[yellow]Found Similar Service Names:")
                console.print(Columns(similar_service_names, equal=True, expand=True))
                return
            else:
                raise typer.Exit(
                    f"Invalid service name: {service}. Please pick a proper one."
                )

        service_node = balcony_aws.get_service_node(service)
        resource_nodes = service_node.get_resource_nodes()

        resource_node_names = []
        for _rn in resource_nodes:
            _rn_name = _rn.name
            # if len(_rn.get_all_required_parameter_names()) >= 2:
            #     _rn_name = f"[bold red]{_rn_name}[/]"
            # elif len(_rn.get_all_required_parameter_names()) >= 1:
            #     _rn_name = f"[bold green]{_rn_name}[/]"
            resource_node_names.append(_rn_name)

        resource_node_name_as_columns = Columns(
            resource_node_names, equal=True, expand=True
        )
        console.print(resource_node_name_as_columns)
        return resource_node_names

    elif service and resource_node:
        # we got both options filled
        service_node = balcony_aws.get_service_node(service)
        resource_nodes = service_node.get_resource_nodes()
        resource_node_names = [_rn.name for _rn in resource_nodes]
        resource_node_obj = service_node.get_resource_node_by_name(resource_node)
        if not resource_node_obj:

            similar_resource_names = ifind_similar_names_in_list(
                resource_node, resource_node_names
            )
            if similar_resource_names:
                console.print("[yellow]Found Similar Resource Node Names:")
                console.print(Columns(similar_resource_names, equal=True, expand=True))
                return
            else:
                raise typer.Exit(
                    f"Invalid Resource Node name: [bold]{resource_node}[/] for Service: [bold]{service}[/]."
                )
        operations_panel = service_node._get_operation_details_panel(
            resource_node_obj.name
        )
        if screen_pager:
            with console.pager(styles=True):
                console.print(operations_panel)
        else:
            console.print(operations_panel)
        return


@app.command("aws")
def aws_main_command(  # noqa
    service: Optional[str] = typer.Argument(
        None,
        show_default=False,
        help="Name of the AWS Service",
        autocompletion=_complete_service_name,
    ),
    resource_node: Optional[str] = typer.Argument(
        None,
        show_default=False,
        help="Name of the AWS Resource Node",
        autocompletion=_complete_resource_node_name,
    ),
    operation: Optional[str] = typer.Argument(
        None,
        show_default=False,
        help="Select a specific operation type. <get, gets, list, describe>",
        autocompletion=_complete_operation_type,
    ),
    patterns: Optional[List[str]] = typer.Option(
        None,
        "--pattern",
        show_default=False,
        help='UNIX pattern matching for generated parameters. Should be quoted. e.g. (--pattern "*prod-*")',
    ),
    jmespath_selector: Optional[str] = typer.Option(
        None,
        "--jmespath-selector",
        "-js",
        show_default=False,
        help="JMESPath query selector to filter resulted data. Visit for tutorial: https://jmespath.org/tutorial.html",
    ),
    debug: bool = typer.Option(False, "--debug", "-d", help="Enable debug messages."),
    formatter: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        show_default=False,
        help="Python f-string expression to generate a string for each item.",
    ),
    list_contents: bool = typer.Option(
        False,
        "--list",
        "-l",
        help="Print the documentation of Service or Resource. Does not make requests.",
    ),
    screen: bool = typer.Option(
        False, "--screen", "-s", help="Open the data on a separate paginator on shell."
    ),
    follow_pagination: bool = typer.Option(
        False,
        "--paginate",
        "-p",
        help="Paginate through the output if the output is truncated, otherwise will only read one page.",
    ),
    output_file: str = typer.Option(
        None,
        "--output",
        "-o",
        show_default=False,
        help="Output JSON file name. If not provided, will print to console.",
    ),
):
    if debug:
        set_log_level_at_runtime(logging.DEBUG)

    if list_contents:
        _list_service_or_resource(service, resource_node, screen_pager=screen)
        raise typer.Exit()

    # warn user if pagination is not set
    if not follow_pagination:
        logger.debug(
            "[underline][yellow bold][WARNING][/] [bold][--paginate, -p][/] option [bold red]is NOT set[/]. You're likely to get incomplete data.[/]"
        )

    if not service and not resource_node:
        # print out resource nodes of this service.
        _list_service_or_resource(
            service, resource_node, screen_pager=screen
        )
        console.print(
            Panel("[bold]Please pick one of the AWS Services", title="[red][bold]ERROR")
        )
        raise typer.Exit()

    if service and not resource_node:
        _list_service_or_resource(
            service, resource_node, screen_pager=screen
        )
        console.print(
            Panel(
                f"[bold]Please pick one of the Resource Nodes from [green]{service}[/] Service",
                title="[red][bold]ERROR",
            )
        )
        raise typer.Exit()

    elif service and resource_node:
        service_node = balcony_aws.get_service_node(service)
        service_reader = service_node.get_service_reader()

        is_operation_selected = operation is not None
        read_data = None
        if not is_operation_selected:
            # read all operations in given resource node
            read_data = service_reader.read_resource_node(
                resource_node,
                match_patterns=patterns,
                follow_pagination=follow_pagination,
            )

        else:  # Operation is selected
            resource_node_obj = service_node.get_resource_node_by_name(resource_node)
            types_to_op_names = resource_node_obj.get_operation_types_and_names()
            supported_operation_types = list(types_to_op_names.keys())
            operation_name = types_to_op_names.get(operation, False)
            if not operation_name:
                console.print(
                    f"[red bold]Given {operation} is not supported by {resource_node}. Try: {supported_operation_types}"
                )
                raise typer.Exit(code=-1)

            read_data = service_reader.read_operation(
                resource_node,
                operation_name,
                match_patterns=patterns,
                refresh=False,
                follow_pagination=follow_pagination,
            )

        if jmespath_selector:

            logger.debug(
                f"Using jmespath selector: {jmespath_selector} to query the returned data."
            )
            read_data = jmespath.search(jmespath_selector, read_data)

        if formatter:
            if read_data:
                formatted_output_list = []
                for r_data in read_data:
                    try:
                        # try to format and add it to output lines list
                        formatted_output_list.append(formatter.format(**r_data))
                    except AttributeError as e:
                        logger.debug(
                            f"Failed to format  [red]{str(e)}[/] with data: {r_data}"
                        )

                if output_file:
                    output_filepath = Path(output_file).resolve()
                    logger.info(f"Saving output to: {output_filepath}")
                    save_str_list_to_output_file(output_filepath, formatted_output_list)
                    raise typer.Exit()
                else:
                    # print to console
                    console.print("\n".join(formatted_output_list), overflow="ignore")
                    raise typer.Exit()
            else:
                logger.debug(
                    f"[red]No data found.[/] Failed to use --format: '{formatter}'"
                )

        if screen:
            with console.pager(styles=True):
                console.print_json(data=read_data, default=str)
        elif output_file:
            output_filepath = Path(output_file).resolve()
            logger.info(f"Saving output to: {output_filepath}")
            save_dict_to_output_file(output_filepath, read_data)
        else:
            console.print_json(data=read_data, default=str)
        raise typer.Exit()


@app.command("clear-cache", help="Clear relations json cache")
def clear_cache_command(
    # service: Optional[str] = typer.Argument(None, show_default='all',
    # help='Name of the Service to clear relation caches of', autocompletion=_complete_service_name),
):
    logger.debug("Deleting relations cache for services:")
    deleted_service_caches = clear_relations_cache()
    for deleted_service in deleted_service_caches:
        logger.info(f"[green]Deleted[/] {deleted_service}")


@app.command("info", help="Information about the AWS Profile and Region currently used", )
def info_command():
    created_session = balcony_aws.boto3_session
    console.print(f"[bold underline]AWS Region:[/] [bold blue]{created_session.region_name}[/]")
    console.print(f"[bold underline]AWS Profile:[/] [bold blue]{created_session.profile_name}")
    console.print()
    console.print("[bold underline]Available Profiles:[/]")

    for available_profile in created_session.available_profiles or []:
        console.print(f"  - [green bold]{available_profile}[/]")
    console.print(textwrap.dedent("""
        [yellow]You can configure the AWS Profile and Region by setting the
        the $[bold]AWS_DEFAULT_REGION[/] and $[bold]AWS_PROFILE[/] environment variables.[/]
    """))


# @app.command('version', help='Show version info' )
# def version_command():
#     from importlib.metadata import version
#     console.print(__package__)
#     version=version(__package__)
#     console.print(version)


def run_app():
    app(prog_name="balcony")


if __name__ == "__main__":
    run_app()



"""
All custom Resource Node classes **must be exported here**.

balcony dynamically imports custom_nodes folder like this:
```python
import custom_nodes
```

When only the module name is given, python will try to import the `__init__.py` file in the package.
So, all custom classes must be exported here to be imported by balcony.
"""

from .codebuild import *
from .ecs import *
from .lambda_functions import *
from .ssm import *
from .ses import *
from .iam import *
from .s3 import *



"""
The general export for the balcony's functionality.

Symbols exported in this file can be used in the following way:

```python
from balcony import BalconyAWS
```

Otherwise you'd need to specify the module name:

```python
from balcony.aws import BalconyAWS
```
"""
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

# TODO: Fix this import mess
try:
    from .aws import BalconyAWS
    from .config import get_rich_console, get_logger
    from .nodes import ServiceNode, ResourceNode
    from .reader import ServiceReader
    from .relations import RelationMap
    from .errors import Error
except ImportError:
    # print('ImportError: balcony/__init__.py')
    from aws import BalconyAWS
    from config import get_rich_console, get_logger
    from nodes import ServiceNode, ResourceNode
    from reader import ServiceReader
    from relations import RelationMap
    from errors import Error



from nodes import ResourceNode
from config import get_logger

logger = get_logger(__name__)

class Template(ResourceNode, service_name="ses", name="Template"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        return [{
            "service_name": "ses",
            "resource_node_name": "Template",
            "required_shape_name": "TemplateName",
            "target_shape_name": "Name",
            "target_shape_type": "string",
            "operation_name": "ListTemplates",
            "target_path": "TemplatesMetadata[*].Name"
        }]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['balcony', 'balcony.custom_nodes']

package_data = \
{'': ['*'], 'balcony': ['custom_yamls/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'boto3>=1.24.80,<2.0.0',
 'inflect>=6.0.0,<7.0.0',
 'jmespath>=1.0.1,<2.0.0',
 'mkdocs-autorefs>=0.4.1,<0.5.0',
 'mkdocs-material>=8.5.7,<10.0.0',
 'mkdocstrings[python]>=0.21.2,<0.22.0',
 'pydantic>=1.10.7,<2.0.0',
 'rich>=13.3.4,<14.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['balcony = balcony.cli:run_app']}

setup_kwargs = {
    'name': 'balcony',
    'version': '0.0.87',
    'description': 'AWS API for humans',
    'long_description': "# balcony - AWS API for humans\n\nbalcony is a read-only CLI tool that simplifies the process of listing all resources on the AWS account.\n\nbalcony dynamically parses AWS SDK(`boto3` library) and analyzes required parameters for each operation. \n\nBy **establishing relations between operations over required parameters**, it's **able to auto-fill** them by reading the related operation beforehand.\n\nBy simply entering their name, balcony enables developers to easily list their AWS resources.\n\n<!-- It uses _read-only_ operations, it does not take any action on the used AWS account. -->\n\n## Installation & Documentation \n\n**[https://oguzhan-yilmaz.github.io/balcony/](https://oguzhan-yilmaz.github.io/balcony/quickstart)**\n\nBalcony's documentation website contains quickstart guide, usage cookbook and more.\n## Features & GIFs\n\n### List available AWS Services \n`balcony aws` to see every AWS service available.\n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/aws-services-list.gif)\n\n\n### List Resource Nodes of an AWS Service \n`balcony aws <service-name>` to see every Resource Node of a service.\n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/resource-node-list.gif)\n\n\n### Reading a Resource Node \n`balcony aws <service-name> <resource-node>` to read operations of a Resource Node.\n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/reading-a-resource-node.gif)\n\n\n### Documentation and Input & Output of Operations\nUse the `--list`, `-l` flag to print the given AWS API Operations documentation, input & output structure. \n \n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/list-option.gif)\n\n\n### Enable Debug messages \nUse the `--debug`, `-d` flag to see what's going on under the hood!\n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/debug-messages.gif)\n",
    'author': 'Oguzhan Yilmaz',
    'author_email': 'oguzhanylmz271@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from utils import icompare_two_camel_case_words
from config import get_rich_console

from typing import List, Union
from botocore.model import Shape, DenormalizedStructureBuilder, OperationModel
from rich.markup import escape
import re
from collections import namedtuple
from rich.tree import Tree

ShapeAndTargetPath = namedtuple("ShapeAndTargetPath", ["shape", "target_path"])

UNWANTED_SHAPE_NAMES = ("String", "DateTime", "Name", "Id", "Arn", "__string")
UNWANTED_SHAPE_NAMES_LOWERED = [_.lower() for _ in UNWANTED_SHAPE_NAMES]
SHAPE_SCALAR_TYPES = DenormalizedStructureBuilder.SCALAR_TYPES
_MAX_ALLOWED_RECURSION = 10
SHAPE_COLLECTION_TYPES = ("structure", "list", "map")
READ_ONLY_VERBS = ("Describe", "List", "Get")
IDENTIFIER_NAMES = (
    "arn",
    "id",
    "name",
    "arns",
    "ids",
    "names",
    "identifier",
    "identifiers",
    "number",
    "url",
)

# shape_resolver can't find the reference of BLACKLISTED_SHAPE_NAMES, so they're ignored
BLACKLISTED_SHAPE_NAMES = (
    "ComponentChildList",
    "FirewallManagerRuleGroups",
    "Rules",
    "HeaderNames",  # noqa
    "ExcludedRules",
    "CountryCodes",
    "CookieNames",
    "TextTransformations",
    "ComponentSummaryList",  # noqa
    "AnomalyMonitors",
    "ConfigurationList",
    "HandshakeResources",
    "JsonPointerPaths",  # noqa
    "AdministrativeActions",
    "DataValueList",
    "ThemeValuesList",
    "Expression",
    "Expressions",
    "CostCategoryRulesList",
    "GetCostAndUsageRequest",
    "GetCostAndUsageWithResourcesRequest",
    "GetAnomaliesRequest",
    "InventoryAggregator",
    "OpsFilter",
    "OpsAggregator",
    "QueryStagePlanNodes",
    "QueryStagePlanNode",
    "QueryStage",
    "ElicitSubSlot",
    "DialogAction"
)  # noqa
# regex expr for removing html caret tags
HTML_CLEANER_REGEX = re.compile("<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});")

console = get_rich_console()


def find_key_in_dict_keys(key: str, dict_keys: Union[list, dict]) -> str:
    """Case insensitive search for a `key` in a `list` or `dict.keys()`.
    Returns the existing key.

    Args:
        key (str): Search case insensively for
        dict_keys (Union[list, dict]): In a list or keys of dictionary

    Returns:
        str: Found key that case insensively matches the given key.
    """
    dict_keys_list = dict_keys
    if type(dict_keys) == dict:
        dict_keys_list = dict_keys.keys()
    for dict_key in dict_keys_list:
        if key.lower() == dict_key.lower():
            return dict_key
    return False


def ifind_key_in_dict_keys(key: str, dict_keys: Union[list, dict]) -> str:
    """Case insensitive search for a `key` in a `list` or `dict.keys()`.

    Args:
        key (str): Search case insensively for
        dict_keys (Union[list, dict]): In a list or keys of dictionary

    Returns:
        str: Found key that case insensively matches the given key.
    """
    dict_keys_list = dict_keys
    if type(dict_keys) == dict:
        dict_keys_list = dict_keys.keys()
    for dict_key in dict_keys_list:
        if icompare_two_camel_case_words(key, dict_key):
            return dict_key
    return False


def get_max_results_value_from_shape(input_shape: Shape) -> int:
    """Finds the `MaxResults` highest value for an input shape.

    Args:
        input_shape (Shape): Input shape of an operation. Generally postfixed with `Request`.

    Returns:
        int: `MaxResults` highest value if found in the Shape definition
    """
    flat_shapes_and_target_paths = (
        flatten_shape_to_its_non_collection_shape_and_target_paths(input_shape)
    )
    flat_members = [_.shape for _ in flat_shapes_and_target_paths]
    for member_shape in flat_members:
        shape_key_name = getattr(member_shape, "key_name", False)
        is_key_name_maxresults = shape_key_name and (
            shape_key_name.lower() == "maxresults"
        )
        shape_name = get_shape_name(member_shape)
        if shape_name.lower() == "maxresults" or is_key_name_maxresults:
            member_shape
            max_value = member_shape.metadata.get("max", False)
            if max_value:
                return max_value
    return False


def get_input_shape(operation_model: OperationModel) -> Shape:
    """Get the input shape of the operation model

    Args:
        operation_model (OperationModel): botocore OperationModel

    Returns:
        Shape: Input Shape of the OperationModel
    """
    return getattr(operation_model, "input_shape", False)


def get_members_shapes(shape: Shape, _recursion_count=0) -> List[Shape]:
    """Get the member shapes of and input or output Shape.

    Args:
        shape (Shape): Shape to get members from

    Returns:
        List[Shape]: Member Shapes, added `key_name` and `parent_name` values to objects.
    """

    found_members_shapes = []

    if _recursion_count >= _MAX_ALLOWED_RECURSION:
        return found_members_shapes

    if not shape:
        return found_members_shapes
    if shape.name in BLACKLISTED_SHAPE_NAMES:
        return found_members_shapes

    if shape.type_name == "structure":
        for shape_key, member in shape.members.items():
            setattr(member, "key_name", shape_key)
            setattr(member, "parent_name", shape.name)
            found_members_shapes.append(member)
    elif shape.type_name == "list":
        only_member = shape.member
        found_members_shapes.append(only_member)
    elif shape.type_name == "map":
        t = get_members_shapes(shape.value, _recursion_count+1)
        return t
    return found_members_shapes


def is_shape_non_collection_type(shape_and_target_path: ShapeAndTargetPath) -> bool:
    """Boolean function to check ShapeAndTargetPath Named Tuple

    Args:
        shape_and_target_path (ShapeAndTargetPath): Named Tuple

    Returns:
        bool: _description_
    """
    shape = shape_and_target_path.shape
    has_key_name = getattr(shape, "key_name", False)
    is_non_collection = shape.type_name not in SHAPE_COLLECTION_TYPES
    return has_key_name and is_non_collection


def _flatten_shape_to_its_members_and_target_paths(
    shape: Shape, target_str: str = ""
) -> List[ShapeAndTargetPath]:
    """Recursive function to get a shape's all members with targetpaths.
    Generates target_str JMESPath selector for each member as it's located in the hierarchy.
    Returns a flat list of (shape, target_path) namedtuples

    Args:
        shape (Shape): botocore shape, possibly output shape of an operation
        target_str (str, optional): Used for keeping track of the target path in recursion.

    Returns:
        List[ShapeAndTargetPath]: List of ShapeAndTargetPath NamedTuple
    """
    result = []
    members = get_members_shapes(shape)
    for member in members:
        member_key_name = getattr(member, "key_name", False)
        new_target_str = target_str
        if member_key_name:
            if target_str == "":
                new_target_str = f"{member_key_name}"
            else:
                new_target_str = f"{target_str}[*].{member_key_name}"
        if member.type_name in (
            "structure",
            "list",
        ):  # TODO:SHAPE_COLLECTION_TYPES-('map',):
            # if the member is a collection type, recurse into it
            inner_list = _flatten_shape_to_its_members_and_target_paths(
                member, new_target_str
            )
            result.extend(inner_list)
        else:
            result.append(ShapeAndTargetPath(member, new_target_str))
    return result


def flatten_shape_to_its_non_collection_shape_and_target_paths(
    shape: Shape,
) -> List[ShapeAndTargetPath]:
    """Return a flat list of shapes all member shapes w/ their JMESPath selector `target_path`

    Args:
        shape (Shape): botocore shape to list its members

    Returns:
        List[ShapeAndTargetPath]: (shape, target_path) custom namedtuple
    """
    # generate all possible members and their target paths
    all_flat_members_and_target_paths = _flatten_shape_to_its_members_and_target_paths(
        shape
    )
    # filter out the collection types beacuse they are not supported by JMESPath
    non_collection_shapes_and_target_paths = list(
        filter(is_shape_non_collection_type, all_flat_members_and_target_paths)
    )
    return non_collection_shapes_and_target_paths


def cleanhtml(raw_html: str) -> str:
    """Removes the HTML tags from given raw_html

    Args:
        raw_html (str): HTML string to clean the tags off of

    Returns:
        str: HTML with tags removed
    """
    cleantext = re.sub(HTML_CLEANER_REGEX, "", raw_html)
    return cleantext


def rich_str_shape(shape: Shape) -> str:
    """Transforms a Shape to rich supported string.

    Args:
        shape (Shape): botocore.model.Shape object.

    Returns:
        str: Rich string for shape.
    """
    key_name = getattr(shape, "key_name", "")
    type_name = str(shape.type_name)
    shape_documentation = cleanhtml(shape.documentation)

    shape_str = (
        f"[blue bold]{key_name}[/] [white]({type_name})[/]: {shape_documentation}"
    )
    if key_name == "":
        lead = ""
        if type_name == "list":
            lead = "["
        elif type_name == "structure":
            lead = "{"
        shape_str = f"[red]{escape(lead)}[/] — [white](({type_name}))[/]: [gray]{shape_documentation}[/]"

    return shape_str


def generate_rich_tree_from_shape(shape: Shape) -> Tree:
    """Genereate a rich Tree containing `shape` and it's members."""
    tree = Tree(rich_str_shape(shape), guide_style="red")

    def _recursive_stringify_shape(shape, node: Tree):

        members = get_members_shapes(shape)
        for member in members:
            member_str = rich_str_shape(member)
            if member.type_name in SHAPE_COLLECTION_TYPES:
                new_node = node.add(member_str)
                _recursive_stringify_shape(member, new_node)
            else:
                node.add(member_str)

    _recursive_stringify_shape(shape, tree)
    return tree


# @lru_cache(maxsize=500) # print(get_shape_name.cache_info())
def get_shape_name(shape: Shape) -> str:
    """Returns the shapes name, using custom set 'key_name' attr

    Args:
        shape (Shape): botocore Shape obj

    Returns:
        str: Name of the shape.
    """
    shape_name = shape.name
    shape_key_name = getattr(shape, "key_name", False)
    if shape_name.lower() in UNWANTED_SHAPE_NAMES_LOWERED:
        # check for name forgotten in shape.serialization if the name is a type name.
        if shape_key_name:
            return shape_key_name
        elif shape.serialization and shape.serialization.get("name", False):
            given_serialization_name = shape.serialization.get("name")
            return given_serialization_name
    if shape_key_name:
        return shape_key_name
    return shape_name


def get_required_parameter_shapes_from_operation_model(
    operation_model: OperationModel,
) -> List[Shape]:
    """Finds required parameter shapes of the operation models input_shape.

    Args:
        operation_model (OperationModel): botocore OperationModel obj of the Operation

    Returns:
        List[Shape]: Required parameter shapes.
    """
    input_shape = get_input_shape(operation_model)
    if not input_shape:
        return []  # there's no input shape
    input_shape_members = get_members_shapes(input_shape)
    input_required_member_names = input_shape.required_members
    if input_required_member_names == []:
        return []

    required_member_shapes = []
    for input_member in input_shape_members:
        input_member_name = get_shape_name(input_member)
        for required_name in input_required_member_names:
            if icompare_two_camel_case_words(required_name, input_member_name):
                required_member_shapes.append(input_member)

    return required_member_shapes
